/*
** Module   : SOAPHeaderImpl.java
** Abstract : Implementation for the SOAP-HEADER resource.
**
** Copyright (c) 2013-2014, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- --------------------------------Description-----------------------------------
** 001 CA  20130216 Created initial version.
** 002 CA  20130221 Added ADM-DATA, UNIQUE-ID support. Fixed resource type.
** 003 OM  20130304 Refactored isValid and isUnknown of WrappedResource to valid and unknown.
** 004 CA  20130603 resourceDelete must return a boolean value.
** 005 CA  20130927 Removed setResourceType, as the type is determined from LegacyResource 
**                  annotation.
** 006 HC  20131215 Implemented ADM-DATA and UNIQUE-ID attribute.
** 007 CA  20140107 Added runtime support.
*/

package com.goldencode.p2j.util;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.*;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.soap.SOAPConstants;

import org.w3c.dom.*;

import com.goldencode.p2j.util.UniqueIdGenerator.IdKind;
import com.goldencode.p2j.xml.*;
import com.goldencode.util.*;

/**
 * A class for implementing the SOAP-HEADER methods and attributes.
 */
public class SOAPHeaderImpl
extends HandleChain
implements SOAPHeader,
           CommonHandleChain
{
   /** Corresponds to ADM-DATA attribute. */
   private final character admData = new character();
   
   /** Corresponds to UNIQUE-ID attribute. */
   private final integer uniqueID = new integer(UniqueIdGenerator.getUniqueId(IdKind.SOAP_HEADER));

   /** All SOAP Header Entries linked with this SOAP Header. */ 
   private final Set<SOAPHeaderEntryImpl> headers = new HashSet<>();
   
   /** 
    * Mapping of each key of a header entry to a DOM XML Node. This is needed because, when a 
    * SOAP Header Entry resource changes its inner DOM XML node, all linked SOAP Header Entries 
    * change it too.
    */ 
   private final Map<Object, Node> entries = new HashMap<>();
   
   /** List allowing fast indexed access to the header entry keys. */
   private final List<Object> entryKeys = new ArrayList<>();

   /** Flag indicating if this resource was deleted or not. */
   private boolean deleted = false;
   
   /**
    * Implicit c'tor.
    */
   public SOAPHeaderImpl()
   {
      super(true, false);
   }

   /**
    * Returns the number of SOAP-HEADER-ENTRYREF object entries attached to the SOAP-HEADER 
    * object.
    * 
    * @return   See above.
    */
   public integer getNumHeaderEntries()
   {
      return new integer(entries.size());
   }

   /**
    * Creates a new entry in a SOAP-header object's list of entries and associates the new entry
    * with the given SOAP-HEADER-ENTRY-REF object.
    * 
    * @param    entryRef
    *           The SOAP Header Entry.

    * @return   <code>true</code> if the operation was possible.
    */
   public logical addHeaderEntry(handle entryRef)
   {
      if (entryRef._isValid() && (entryRef.getResource() instanceof SOAPHeaderEntryImpl))
      {
         SOAPHeaderEntryImpl entry = (SOAPHeaderEntryImpl) entryRef.getResource();
         addHeaderEntry(entry);

         return new logical(true);
      }

      SOAPFactory.throwError("ADD-HEADER-ENTRY", "SOAP-HEADER-ENTRY");

      return new logical(false);
   }


   /**
    * Get the header entry at the given index and assign it to the given handle.
    * <p>
    * This method returns a logical even though the Progress 4GL documentation explicitly states 
    * otherwise. Actual working code has proven that the original documentation is incorrect. 
    * This implementation matches the actual runtime behavior of the 4GL.
    * 
    * @param    target
    *           The handle where the SOAP-HEADER-ENTRY instance will be saved.
    * @param    idx
    *           The index of the SOAP-HEADER-ENTRY instance to retrieve. 4GL allows only integer
    *           and int64 values to be passed to this parameter.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical getHeaderEntry(handle target, long idx)
   {
      return getHeaderEntry(target, new int64(idx));
   }

   /**
    * Get the header entry at the given index and assign it to the given handle.
    * <p>
    * This method returns a logical even though the Progress 4GL documentation explicitly states 
    * otherwise. Actual working code has proven that the original documentation is incorrect. 
    * This implementation matches the actual runtime behavior of the 4GL.
    * 
    * @param    target
    *           The handle where the SOAP-HEADER-ENTRY instance will be saved.
    * @param    idx
    *           The index of the SOAP-HEADER-ENTRY instance to retrieve. 4GL allows only integer
    *           and int64 values to be passed to this parameter.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical getHeaderEntry(handle target, integer idx)
   {
      return getHeaderEntry(target, new int64(idx));
   }

   /**
    * Get the header entry at the given index and assign it to the given handle.
    * <p>
    * This method returns a logical even though the Progress 4GL documentation explicitly states 
    * otherwise. Actual working code has proven that the original documentation is incorrect. 
    * This implementation matches the actual runtime behavior of the 4GL.
    * 
    * @param    target
    *           The handle where the SOAP-HEADER-ENTRY instance will be saved.
    * @param    idx
    *           The index of the SOAP-HEADER-ENTRY instance to retrieve. 4GL allows only integer
    *           and int64 values to be passed to this parameter.
    * 
    * @return   <code>true</code> if the operation was possible.
    */
   public logical getHeaderEntry(handle target, int64 idx)
   {
      if (target._isValid() && (target.getResource() instanceof SOAPHeaderEntryImpl))
      {
         if (idx.isUnknown() || idx.intValue() < 1 || idx.intValue() > entries.size())
         {
            // no error is shown or registered, and true is returned.
            return new logical(true);
         }
         
         // locate the key
         Object key = entryKeys.get(idx.intValue() - 1);

         // this returns really a new resource... it has distinct ADM-DATA/UNIQUE-ID/PRIVATE-DATA,
         // but they share the inner X-NODEREF resource.
         SOAPHeaderEntryImpl targetEntry = (SOAPHeaderEntryImpl) target.getResource();
         
         targetEntry.setParent(this, key);
         
         // register this header, so it can be detached if the soap-header gets deleted
         this.headers.add(targetEntry);

         return new logical(true);
      }

      SOAPFactory.throwError("GET-HEADER-ENTRY", "SOAP-HEADER-ENTRY");
      
      return new logical(false);
   }
   
   /**
    * Reports if this object is valid for use.  
    *
    * @return   <code>true</code> if we are valid (can be used).
    */
   public boolean valid()
   {
      return !deleted;
   }

   /**
    * Get the value of the ADM-DATA attribute.
    * 
    * @return   See above.
    */
   @Override
   public character getADMData()
   {
      return new character(this.admData);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(String value)
   {
      this.admData.assign(value);
   }

   /**
    * Set the value of the ADM-DATA attribute.
    * 
    * @param    value
    *           The new value.
    */
   @Override
   public void setADMData(character value)
   {
      this.admData.assign(value);
   }

   /**
    * Gets the the unique ID number associated to this object by the underlying system. Not the 
    * same as the handle value itself.
    * 
    * @return   The integer number of the children.
    */
   @Override
   public integer getUniqueID()
   {
      return new integer(this.uniqueID);
   }

   /**
    * Delete the resource.
    * 
    * @return   <code>true</code> if the resource was deleted.
    */
   @Override
   protected boolean resourceDelete()
   {
      // do not delete the soap header entries, but unlink them.
      for (SOAPHeaderEntryImpl header : headers)
      {
         header.setParent(null, null);
      }
      
      // clear all references
      entries.clear();
      entryKeys.clear();
      headers.clear();
      
      this.deleted = true;

      return true;
   }

   /**
    * Check if this resource supports the NEXT-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasNextSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the PREV-SIBLING attribute.
    * 
    * @return   Always <code>false</code>.
    */
   @Override
   protected boolean hasPrevSibling()
   {
      return false;
   }
   
   /**
    * Check if this resource supports the NAME attribute in read-only mode.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   protected boolean hasNameReadOnly()
   {
      return true;
   }

   /**
    * Removes the specified SOAP-Header object from the list of header entries.
    * 
    * @param    entry
    *           The SOAP Header Entry.
    * @param    key
    *           The key to the DOM XML Node referenced by this header entry.
    */
   void removeHeaderEntry(SOAPHeaderEntryImpl entry, Object key)
   {
      // cleanup after this key
      entries.remove(key);
      entryKeys.remove(key);
      headers.remove(entry);
      
      Iterator<SOAPHeaderEntryImpl> itr = headers.iterator();
      while (itr.hasNext())
      {
         SOAPHeaderEntryImpl header = itr.next();

         if (header.getHeaderKey() == key)
         {
            // detach and remove them
            header.setParent(null, null);
            itr.remove();
         }
      }
   }
   
   /**
    * Creates a new entry in a SOAP-header object's list of entries and associates the new entry
    * with the given SOAP-HEADER-ENTRY-REF object.
    * 
    * @param    entry
    *           The SOAP Header Entry.
    */
   void addHeaderEntry(SOAPHeaderEntryImpl entry)
   {
      Object key = new Object();

      entry.setParent(this, key);

      entries.put(key, null);
      entryKeys.add(key);
      
      headers.add(entry);
   }

   /**
    * Get the string representation of the XML payload associated with this soap header.
    * <p>
    * If there are no header entries, return {@code null}.
    * 
    * @return   See above.
    */
   String asString()
   {
      try
      {
         Document doc =  XmlHelper.newDocument();

         // create a HEADER element node, linked with the SOAP 1.1. namespace; to this, each 
         // header entry DOM XML node will be added.
         Element root = doc.createElementNS(SOAPConstants.URI_NS_SOAP_1_1_ENVELOPE, "header");
         doc.appendChild(root);

         boolean hadHeaders = false;
         for (Object key : entryKeys)
         {
            Node entry = entries.get(key);

            if (entry == null)
            {
               // TODO: what if a key has no DOM XML node? this is possible, because the DOM XML
               // node is not created until SOAP-HEADER-ENTRYREF:SET-NODE or SET-SERIALIZED is
               // called
               continue;
            }
            
            Element clone = (Element) doc.importNode(entry, true);
            
            String actor = (String) entry.getUserData(SOAPHeaderEntryImpl.ACTOR);
            if (actor != null)
            {
               clone.setAttributeNS(SOAPConstants.URI_NS_SOAP_1_1_ENVELOPE, "actor", actor);
            }
            
            Boolean mustUnderstand =
               (Boolean) entry.getUserData(SOAPHeaderEntryImpl.MUST_UNDERSTAND);
            if (mustUnderstand != null && mustUnderstand)
            {
               clone.setAttributeNS(SOAPConstants.URI_NS_SOAP_1_1_ENVELOPE, "mustUnderstand", "1");
            }
            
            root.appendChild(clone);
            
            hadHeaders = true;
         }
         
         String headers = null;

         if (hadHeaders)
         {
            ByteArrayOutputStream output = new ByteArrayOutputStream();
            XmlHelper.write(doc, output, false, true);
            
            headers = output.toString().trim();
         }
         
         return headers;
      }
      catch (ParserConfigurationException | IOException e)
      {
         throw new RuntimeException("Problems serializing SOAP Headers:", e);
      }
   }
   
   /**
    * Get the DOM XML node associated with the given key.
    * 
    * @param    key
    *           The key to an entry in the {@link #entries} map.
    *           
    * @return   The DOM XML node for the given key.
    * 
    * @throws   IllegalArgumentException
    *           If the received key is {@code null} or is not part of the {@link #entries} map's 
    *           keys.
    */
   Node getHeaderNode(Object key)
   {
      if (key != null && entries.containsKey(key))
      {
         return entries.get(key);
      }
      else
      {
         throw new IllegalArgumentException("Received an unknown SOAP Header Entry key!");
      }
   }
   
   /**
    * Associated a new DOM XML node with the given header entry key.
    * 
    * @param    key
    *           The key to an entry in the {@link #entries} map.
    * @param    node
    *           The new DOM XML node for the given key.
    *           
    * @throws   IllegalArgumentException
    *           If the received key is {@code null} or is not part of the {@link #entries} map's 
    *           keys.
    */
   void setHeaderNode(Object key, Node node)
   {
      if (key != null && entries.containsKey(key))
      {
         this.entries.put(key, node);
      }
      else
      {
         throw new IllegalArgumentException("Received an unknown SOAP Header Entry key!");
      }
   }
}
