/*
** Module   : WebServiceHelper.java
** Abstract : A helper class to implement web service calls. 
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20131220 Created initial version.
*/

package com.goldencode.p2j.util;

import java.util.*;
import java.util.logging.*;

import javax.wsdl.*;
import javax.wsdl.extensions.*;
import javax.wsdl.extensions.schema.*;
import javax.wsdl.extensions.soap.*;
import javax.wsdl.extensions.soap12.*;
import javax.wsdl.factory.*;
import javax.wsdl.xml.*;
import javax.xml.namespace.*;
import javax.xml.stream.*;

import org.apache.axiom.om.*;
import org.apache.axiom.om.util.*;
import org.apache.axiom.soap.SOAPHeaderBlock;
import org.apache.axiom.soap.SOAPHeader;
import org.apache.axis2.*;
import org.apache.axis2.addressing.EndpointReference;
import org.apache.axis2.client.*;
import org.apache.axis2.context.MessageContext;
import org.apache.axis2.wsdl.WSDLConstants;
import org.apache.ws.commons.schema.*;
import org.w3c.dom.*;

import com.goldencode.p2j.main.*;
import com.goldencode.p2j.util.ControlFlowOps.ArgValidationErrors;
import com.goldencode.util.*;

/**
 * Provides a collection of APIs for establishing Progress-style connections to web-services and
 * performing Progress-style invocation of service operations.
 * <p>
 * This class allows connections only to SOAP 1.1 services, to conform with Progress runtime.
 * <p>
 * Rules describing the validation of the web-service connect options and validation of the WSDL 
 * file:
 * <ol>
 * <li>the target WSDL file must be a well-formed WSDL file; local or remote URLs can be used.
 * </li>
 * <li>if <code>-Service</code> or <code>-Port</code>is specified, then <code>-Binding</code> and
 *     <code>-SOAPEndpoint</code> must not be specified (and viceversa).</li>
 * <li>if <code>-Binding</code> is specified, then <code>-SOAPEndpoint</code> must be specified.
 * </li>
 * <li>if <code>-TargetNamespace</code> is specified, it must match the one at the WSDL document.
 * </li>
 * <li>if <code>-Service</code> is specified, then:
 *  <ul>
 *   <li>it must match (case-sensitively) a service defined in the WSDL document.</li>
 *   <li>if <code>-Port</code> is specified, then the port must be matched with one specified at 
 *       the service in the WSDL document and it must have an associated SOAP 1.1. binding.</li>
 *   <li>if no <code>-Port</code> is specified, then there must exist only one SOAP 1.1. 
 *       <code>(port, binding)</code> pair at the service configuration.</li>
 *  </ul>
 * </li>
 * <li>if <code>-Binding</code> is specified, then:
 *  <ul>
 *   <li>it must be defined in the WSDL document.</li>
 *   <li>it must be for SOAP 1.1.</li>
 *   <li>there must exist a service configured for this binding and its associated port.</li>
 *  </ul>
 * </li>
 * <li>else, if no <code>-Binding</code> or <code>-Service</code> is specified, then:
 *  <ul>
 *   <li>there must exist exactly only on service definied in the document.</li>
 *   <li>the service must have associated only one SOAP 1.1. port.</li>
 *  </ul>
 * </li>
 * </ol>
 */
public class WebServiceHelper
extends ServerHelper
{
   /** Logger */
   private static final Logger LOG = LogHelper.getLogger(WebServiceHelper.class.getName());

   /** Constant identifying the '-WSDL' connect option. */
   static final String CONNECT_OPTION_WSDL = "-WSDL";

   /** Constant identifying the '-WSDLUserid' connect option. */
   static final String CONNECT_OPTION_WSDL_USERID = "-WSDLUserid";

   /** Constant identifying the '-WSDLPassword' connect option. */
   static final String CONNECT_OPTION_WSDL_PASSWORD = "-WSDLPassword";

   /** Constant identifying the '-Service' connect option. */
   static final String CONNECT_OPTION_SERVICE = "-Service";

   /** Constant identifying the '-ServiceNamespace' connect option. */
   static final String CONNECT_OPTION_SERVICE_NAMESPACE = "-ServiceNamespace";

   /** Constant identifying the '-Port' connect option. */
   static final String CONNECT_OPTION_PORT = "-Port";

   /** Constant identifying the '-Binding' connect option. */
   static final String CONNECT_OPTION_BINDING = "-Binding";

   /** Constant identifying the '-BindingNamespace' connect option. */
   static final String CONNECT_OPTION_BINDING_NAMESPACE = "-BindingNamespace";

   /** Constant identifying the '-SOAPEndpoint' connect option. */
   static final String CONNECT_OPTION_SOAP_ENDPOINT = "-SOAPEndpoint";

   /** Constant identifying the '-SOAPEndpointUserid' connect option. */
   static final String CONNECT_OPTION_SOAP_ENDPOINT_USER = "-SOAPEndpointUserid";

   /** Constant identifying the '-SOAPEndpointPassword' connect option. */
   static final String CONNECT_OPTION_SOAP_ENDPOINT_PASSWORD = "-SOAPEndpointPassword";

   /** Constant identifying the '-TargetNamespace' connect option. */
   static final String CONNECT_OPTION_TARGET_NAMESPACE = "-TargetNamespace";

   /** Constant identifying the '-maxConnections' connect option. */
   static final String CONNECT_OPTION_MAX_CONNECTIONS = "-maxConnections";

   /** Constant identifying the '-connectionLifetime' connect option. */
   static final String CONNECT_OPTION_CONNECTION_LIFETIME = "-connectionLifetime";

   /** Constant identifying the '-nosessionreuse' connect option. */
   static final String CONNECT_OPTION_NO_SESSION_REUSE = "-nosessionreuse";

   /** Constant identifying the '-nohostverify' connect option. */
   static final String CONNECT_OPTION_NO_HOST_VERIFY = "-nohostverify";

   /** Constant identifying the '-pf' connect option. */
   static final String CONNECT_OPTION_PF = "-pf";

   /** The set of web-service connect options used by the legacy CONNECT method. */
   private static final Set<String> knownConnectOptions;
   
   static
   {
      knownConnectOptions = new HashSet<String>();
      knownConnectOptions.add(CONNECT_OPTION_WSDL);
      knownConnectOptions.add(CONNECT_OPTION_WSDL_USERID);
      knownConnectOptions.add(CONNECT_OPTION_WSDL_PASSWORD);
      knownConnectOptions.add(CONNECT_OPTION_SERVICE);
      knownConnectOptions.add(CONNECT_OPTION_SERVICE_NAMESPACE);
      knownConnectOptions.add(CONNECT_OPTION_PORT);
      knownConnectOptions.add(CONNECT_OPTION_BINDING);
      knownConnectOptions.add(CONNECT_OPTION_BINDING_NAMESPACE);
      knownConnectOptions.add(CONNECT_OPTION_SOAP_ENDPOINT);
      knownConnectOptions.add(CONNECT_OPTION_SOAP_ENDPOINT_USER);
      knownConnectOptions.add(CONNECT_OPTION_SOAP_ENDPOINT_PASSWORD);
      knownConnectOptions.add(CONNECT_OPTION_TARGET_NAMESPACE);
      knownConnectOptions.add(CONNECT_OPTION_CONNECTION_LIFETIME);
      knownConnectOptions.add(CONNECT_OPTION_MAX_CONNECTIONS);
      knownConnectOptions.add(CONNECT_OPTION_NO_SESSION_REUSE);
      knownConnectOptions.add(CONNECT_OPTION_NO_HOST_VERIFY);
      knownConnectOptions.add(CONNECT_OPTION_PF);
   }
   
   /** The options used to established the connection to the web server. */
   private final WebServiceConnectOptions options;
   
   /** The parsed WSDL document. */
   private final Definition wsdlDef;
   
   /** The service used by this connection. */
   private final Service service;
   
   /** The target SOAP endpoint to which requests are sent. */
   private final String soapEndpoint;
   
   /** The port-type used by this connection. */
   private final PortType portType;
   
   /** A SOAP factory to create message data. */
   private final org.apache.axiom.soap.SOAPFactory soapFactory = 
                                                      OMAbstractFactory.getSOAP11Factory();

   private final Map<QName, QName[][]> msgParams = new HashMap<>();

   private List<XmlSchema> xmlSchemas = null;
   

   /** The counter for async requests.*/
   private int nextAsyncReqId = 0;

   /**
    * Instantiate a new helper.
    * 
    * @param    options
    *           The options used to established the connection to the web server.
    * @param    wsdlDef
    *           The parsed WSDL document.
    * @param    service
    *           The service used by this connection.
    * @param    soapEndpoint
    *           The target SOAP endpoint to which requests are sent.
    * @param    portType
    *           The port-type used by this connection.
    */
   private WebServiceHelper(WebServiceConnectOptions options,
                            Definition               wsdlDef,
                            Service                  service,
                            String                   soapEndpoint,
                            PortType                 portType)
   {
      this.options      = options;
      this.wsdlDef      = wsdlDef;
      this.service      = service;
      this.soapEndpoint = soapEndpoint;
      this.portType     = portType;
   }

   /**
    * Establish a new connection using the given options.
    * <p>
    * The {@link #CONNECT_OPTION_MAX_CONNECTIONS}, {@link #CONNECT_OPTION_CONNECTION_LIFETIME}, 
    * {@link #CONNECT_OPTION_NO_SESSION_REUSE} and {@link #CONNECT_OPTION_NO_HOST_VERIFY} options
    * are not supported.
    * 
    * @param    options
    *           The options used to established the connection to the web server.
    *           
    * @return   A new {@link WebServiceHelper} instance or <code>null</code> if no connection
    *           was possible.  Check the {@link ErrorManager} for details about any errors.
    */
   @SuppressWarnings("unchecked")
   public static WebServiceHelper connect(WebServiceConnectOptions options)
   {
      // step 1. check if the WSDL document is available and can be parsed properly. use the
      //         specified [WSDLUserid [WSDLPassword]]
      Definition wsdlDef = null;
      try
      {
         // TODO: use the WSDL userId and password
         WSDLReader reader = WSDLFactory.newInstance().newWSDLReader();
         
         // TODO: what if the WSDL points to a local file, or to a host which is accessible from
         // the client, but not from the server?
         wsdlDef = reader.readWSDL(options.getWsdlDocument());
      }
      catch (WSDLException e)
      {
         String msg = "Error loading WSDL document  %s :  %s";
         msg = String.format(msg, options.getWsdlDocument(), e.getMessage());
         ErrorManager.recordOrShowError(11748, msg, false, false, false);
         
         if (LOG.isLoggable(Level.SEVERE))
         {
            LOG.log(Level.SEVERE, msg, e);
         }

         return null;
      }
      
      // TODO: validate -ServiceNamespace and -BindingNamespace
      
      // step2a. check if both service and binding are specified
      if ((options.getService() != null || options.getPort() != null) &&
          (options.getBinding() != null || options.getSoapEndpoint() != null))
      {
         final int[] errCodes;
         final String[] errMsgs;
         
         final String msg1 =
            "If -Service and/or -Port are specified in the SERVER:CONNECT " +
            "connection-parameters string, neither -Binding nor -SOAPEndpoint is valid";
         final String msg2 =
            "If '-Binding' is specified in the SERVER:CONNECT connection-parameters string, " +
            "'-SOAPEndPoint' must also be specified";
         
         if (options.getSoapEndpoint() == null && options.getBinding() != null)
         {
            errCodes = new int[] { 11499, 11501 };
            errMsgs = new String[] { msg1, msg2 };
         }
         else
         {
            errCodes = new int[] { 11499 };
            errMsgs = new String[] { msg1 };
         }

         ErrorManager.recordOrShowError(errCodes, errMsgs, false, false, false);
         
         return null;
      }

      // step 2b. if -Binding is specified, -SoapEndpoint must be specified
      if (options.getBinding() != null && options.getSoapEndpoint() == null)
      {
         final String msg =
            "If '-Binding' is specified in the SERVER:CONNECT connection-parameters string, " +
            "'-SOAPEndPoint' must also be specified";
         ErrorManager.recordOrShowError(11501, msg, false, false, false);
         
         return null;
      }
      
      String ns = wsdlDef.getTargetNamespace();
      
      String soapEndpoint = null;
      PortType portType   = null;
      Service service     = null;

      // step 3. only service or binding must be specified
      if (options.getService() != null)
      {
         // step 3a. check if specified [Service [ServiceNamespace]] [Port] are valid
         service = wsdlDef.getService(new QName(ns, options.getService()));
         
         if (service == null)
         {
            final String msg = "The Service name %s is not in the WSDL document";
            ErrorManager.recordOrShowError(11736, String.format(msg, options.getService()),
                                           false, false, false);
            
            return null;
         }
         
         String port = options.getPort();
         final Port servicePort;
         if (port != null)
         {
            servicePort = service.getPort(port);
         }
         else
         {
            servicePort = getImplicitSoapPort(service.getPorts().values());
            
            if (servicePort == null)
            {
               noServicesError();
               return null;
            }
         }
         
         // validate the port
         Binding binding = (servicePort == null ? null : servicePort.getBinding());
         if (servicePort == null ||
             binding == null     ||
             getExtension(binding.getExtensibilityElements(), SOAPBinding.class) == null)
         {
            final String msg = "The Port name %s is not in the WSDL document";
            ErrorManager.recordOrShowError(11742, String.format(msg, port), false, false, false);
            
            return null;
         }

         portType = binding.getPortType();

         SOAPAddress soapAddress = getExtension(servicePort.getExtensibilityElements(), 
                                                SOAPAddress.class);
         soapEndpoint = (soapAddress == null ? null : soapAddress.getLocationURI());
      }
      else if (options.getBinding() != null)
      {
         // step 3b. check if specified [[Binding [BindingNamespace]] [SOAPEndpoint]] are valid
         Binding binding = wsdlDef.getBinding(new QName(ns, options.getBinding()));
         if (binding != null &&
             getExtension(binding.getExtensibilityElements(), SOAP12Binding.class) != null)
         {
            final String msg = "The %s:%s Binding is not supported, it must be a SOAP 1.1 binding";
            ErrorManager.recordOrShowError(12100, String.format(msg, ns, options.getBinding()),
                                           false, false, false);

            return null;
         }
         
         if (binding == null ||
             getExtension(binding.getExtensibilityElements(), SOAPBinding.class) == null)
         {
            final String msg = "The %s:%s Binding name cannot be found in the WSDL document";
            ErrorManager.recordOrShowError(11740, String.format(msg, ns, options.getBinding()),
                                           false, false, false);

            return null;
         }
         
         // compute the port type
         portType = binding.getPortType();

         // determine the target service, based on the binding and associated port type
         Iterator<Service> itrSrv = wsdlDef.getAllServices().values().iterator();
         while (itrSrv.hasNext())
         {
            Service srv = itrSrv.next();
            Iterator<Port> itrPorts = srv.getPorts().values().iterator();
            while (itrPorts.hasNext())
            {
               Port port = itrPorts.next();
               if (port.getBinding().equals(binding))
               {
                  service = srv;
                  break;
               }
            }
         }
         
         if (service == null)
         {
            final String msg = 
               "Internal error locating Service and Port for Binding %s:%s (11739)";
            ErrorManager.recordOrShowError(11739, String.format(msg, ns, options.getBinding()),
                                           false, false, false, true);

            return null;
         }
         
         soapEndpoint = options.getSoapEndpoint(); // was already validated

         // TODO: add the soapEndpointUserId and soapEnpointPassword to the endpoint URL
         
         // TODO: check if the endpoint is a valid URL
      }
      else
      {
         // no explicit binding or service, get the default one.

         Map services = wsdlDef.getServices();
         if (services.size() != 1)
         {
            if (services.size() == 0)
            {
               noServicesError();
            }
            else
            {
               final String msg =
                  "A specific Service name must be specified from the WSDL document";
               ErrorManager.recordOrShowError(11737, msg, false, false, false);
            }
            
            return null;
         }
         
         service = (Service) services.values().iterator().next();
         Port servicePort = getImplicitSoapPort(service.getPorts().values());
         
         if (servicePort == null)
         {
            noServicesError();
            return null;
         }
         
         Binding binding = servicePort.getBinding();
         portType = binding.getPortType();
         
         SOAPAddress soapAddress = getExtension(binding.getExtensibilityElements(), 
                                                SOAPAddress.class);
         soapEndpoint = (soapAddress == null ? null : soapAddress.getLocationURI());
      }
      
      // step 4. check if the targetNamespace matches the WSDL document's targetNamespace
      if (options.getTargetNamespace() != null && 
          !options.getTargetNamespace().equals(wsdlDef.getTargetNamespace()))
      {
         final String msg =
            "The TargetNamespace option %s does not match the WSDL's target namespace (11750)";
         
         ErrorManager.recordOrShowError(11750, String.format(msg, options.getTargetNamespace()),
                                        false, false, false, true);
         
         return null;
      }
      
      // everything is valid, instantiate a new helper and save the:
      // - connect options
      // - parsed WSDL document
      // - soapEndpoint
      // - exposed port type (which defines the operations)

      return new WebServiceHelper(options, wsdlDef, service, soapEndpoint, portType);
   }
   
   /**
    * Establish a new web-service connection using the connect options in the given map.
    * 
    * @param    opts
    *           A map with the connect options.
    *           
    * @return   A new {@link WebServiceHelper} instance or <code>null</code> if no connection
    *           was possible.  Check the {@link ErrorManager} for details about any errors.
    */
   static ServerHelper connect(Map<String, String> opts)
   {
      // wsdl target is required
      String wsdlTarget = ConnectHelper.getString(opts, CONNECT_OPTION_WSDL, null);
      // wsdl user/passwd are optional
      String wsdlUserid = ConnectHelper.getString(opts, CONNECT_OPTION_WSDL_USERID, null);
      String wsdlPassword =
         (wsdlUserid == null) ? null : ConnectHelper.getString(opts,
                                                               CONNECT_OPTION_WSDL_PASSWORD,
                                                               "");
      
      // -Server and -Binding groups are mutually exclusive, but are optional
      
      String service = ConnectHelper.getString(opts, CONNECT_OPTION_SERVICE, null);
      String serviceNamespace =
         ConnectHelper.getString(opts, CONNECT_OPTION_SERVICE_NAMESPACE, null);
      String port = ConnectHelper.getString(opts, CONNECT_OPTION_PORT, null);
      
      String binding = ConnectHelper.getString(opts, CONNECT_OPTION_BINDING, null);
      String bindingNamespace =
         ConnectHelper.getString(opts, CONNECT_OPTION_BINDING_NAMESPACE, null);
      String soapEndpoint = ConnectHelper.getString(opts, CONNECT_OPTION_SOAP_ENDPOINT, null);
      
      // soap user/passwd are optional
      String soapEndpointUserid =
         ConnectHelper.getString(opts, CONNECT_OPTION_SOAP_ENDPOINT_USER, null);
      String soapEndpointPassword = (soapEndpointUserid == null)
           ? null
           : ConnectHelper.getString(opts, CONNECT_OPTION_SOAP_ENDPOINT_PASSWORD, "");
      
      // target NS is optional, used to validate the WSDL
      String targetNamespace =
         ConnectHelper.getString(opts, CONNECT_OPTION_TARGET_NAMESPACE, null);
      
      int maxConnections = ConnectHelper.getInt(opts, CONNECT_OPTION_MAX_CONNECTIONS, 0);
      int connectionLifetime =
         ConnectHelper.getInt(opts, CONNECT_OPTION_CONNECTION_LIFETIME, 300);
      
      // related to ssl connections
      boolean noSessionReuse = opts.containsKey(CONNECT_OPTION_NO_SESSION_REUSE);
      boolean noHostVerify = opts.containsKey(CONNECT_OPTION_NO_HOST_VERIFY);
               
      WebServiceConnectOptions options = new WebServiceConnectOptions(wsdlTarget);
      options.setWsdlUserid(wsdlUserid);
      options.setWsdlPassword(wsdlPassword);
      options.setService(service);
      options.setServiceNamespace(serviceNamespace);
      options.setPort(port);
      options.setBinding(binding);
      options.setBindingNamespace(bindingNamespace);
      options.setSoapEndpoint(soapEndpoint);
      options.setSoapEndpointUserid(soapEndpointUserid);
      options.setSoapEndpointPassword(soapEndpointPassword);
      options.setTargetNamespace(targetNamespace);
      options.setMaxConnections(maxConnections);
      options.setConnectionLifetime(connectionLifetime);
      options.setNoSessionReuse(noSessionReuse);
      options.setNoHostVerify(noHostVerify);
   
      return connect(options);
   }

   /**
    * Get the set of known options for web-service connections.
    * 
    * @return   See above.
    */
   static Set<String> getKnownConnectOptions()
   {
      return knownConnectOptions;
   }

   /**
    * Search the through the list of {@link ExtensibilityElement}s for one matching the give
    * type.
    * 
    * @param    extensions
    *           The list of extensibility elements.
    * @param    type
    *           The type to be matched.
    *           
    * @return   The found element or <code>null</code> if not found.
    */
   @SuppressWarnings("unchecked")
   private static <T> T getExtension(List<?> extensions, Class<T> type)
   {
      if (extensions == null)
      {
         return null;
      }
      
      Iterator<ExtensibilityElement> extElIter =
         (Iterator<ExtensibilityElement>) extensions.iterator();
      while (extElIter.hasNext())
      {
         ExtensibilityElement extEl = extElIter.next();
         if (type.isAssignableFrom(extEl.getClass()))
         {
            return (T) extEl;
         }
      }
      
      return null;
   }

   /**
    * Get the implicit port from the given collection of ports.  The found {@link Port} must be
    * associated with a {@link Binding} with a {@link SOAPBinding SOAP 1.1. extensibility}.  If
    * more than one port matches or no port found, return <code>null</code>.
    * 
    * @param    ports
    *           The {@link Port} instances to be searched.
    *           
    * @return   The found {@link Port} following the above rules.
    */
   @SuppressWarnings("unchecked")
   private static Port getImplicitSoapPort(Collection<?> ports)
   {
      Iterator<Port> portIter = (Iterator<Port>) ports.iterator();
      Port soapPort = null;
      while (portIter.hasNext())
      {
         Port port = portIter.next();
         Binding binding = port.getBinding();
         
         if (getExtension(binding.getExtensibilityElements(), SOAPBinding.class) != null)
         {
            if (soapPort != null)
            {
               // more than one SOAP port found, return "not found"
               return null;
            }
            
            soapPort = port;
         }
      }
      
      return soapPort;
   }
   
   /**
    * Show a specific 4GL-style error meaning that no service could be resolved from the WSDL
    * document.
    */
   private static void noServicesError()
   {
      ErrorManager.recordOrShowError(11735, "No Services located in the WSDL document",
                                     false, false, false);
   }

   /**
    * Run the port type with the given name, only if the name matches (case-insensitively) with 
    * the target {@link #portType port type} determined during connection.
    * 
    * @param    name
    *           The port type name.
    *           
    * @return   A {@link PortTypeWrapper} instance associated with the port type or 
    *           <code>null</code> if the port type name does not match.
    */
   public PersistentProcedure runPort(character name)
   {
      // validate the port type
      String portTypeName = name.toStringMessage();

      // search is done case-insensitive
      if (!portType.getQName().getLocalPart().equalsIgnoreCase(portTypeName))
      {
         final int[] errCodes = { 11754, 11755 };
         final String[] errMsgs =
         {
           String.format("Failure in initializing Port Type %s (11754)", portTypeName),
           String.format("Invalid WSDL PortType name %s (11755)", portTypeName)
         };
         ErrorManager.recordOrThrowError(errCodes, errMsgs, false, true);
         
         return null;
      }
      
      return new PortTypeWrapper(portTypeName, this.getServer());
   }
   
   /**
    * Get the sub-type of this web service connection.
    * 
    * @return   Always the <code>WEBSERVICE</code> string.
    */
   @Override
   public String getSubType()
   {
      return "WEBSERVICE";
   }

   /**
    * Get the connection ID of this web service connection.
    * 
    * @return   The {@link #soapEndpoint}.
    */
   @Override
   public String getConnectionID()
   {
      return this.soapEndpoint;
   }

   /**
    * Check if this is a web service.
    * 
    * @return   Always <code>true</code>.
    */
   @Override
   public boolean isWebService()
   {
      return true;
   }

   /**
    * Check if this is a session-free connection.
    * 
    * @return   Always <code>true</code> for web service connections.
    */
   @Override
   public boolean isSessionFree()
   {
      return true;
   }

   /**
    * Get the name of this server.
    * 
    * @return   Always the connected {@link #soapEndpoint}.
    */
   @Override
   public String getServerName()
   {
      return soapEndpoint;
   }

   /**
    * Disconnect this server.
    * <p>
    * This is a no-op for web-services.
    */
   @Override
   public void disconnect()
   {
      // this is a no-op
   }

   /**
    * Send a STOP condition to the specified async request.
    * 
    * @param    requestId
    *           The request where to send a STOP condition.
    */
   @Override
   public void sendStop(int requestId)
   {
      // TODO: implement this
   }

   /**
    * Invoke a web service operation. 
    * 
    * @param    name
    *           The name for the web service operation.
    * @param    h
    *           The port-type procedure in which to invoke the web service operation. May not
    *           be <code>null</code>.
    * @param    function
    *           <code>true</code> if this is a function call.
    * @param    dynamicFunction
    *           <code>true</code> if this is a DYNAMIC-FUNCTION call.
    * @param    superCall
    *           <code>true</code> if this is a RUN SUPER or SUPER() call.
    * @param    transactionDistinct
    *           Flag indicating if the TRANSACTION DISTINCT clause is in effect.
    * @param    modes
    *           A string representation of the modes of each parameter. May not be 
    *           <code>null</code>.
    * @param    args
    *           The procedure's arguments.
    * 
    * @return   The result of this procedure/function invocation or <code>null</code> if the
    *           invocation could not be performed.
    * 
    * @throws   IllegalArgumentException
    *           If this is a super call, the argument modes are not specified or the passed 
    *           procedure handle is <code>null</code> or not a {@link PortTypeWrapper} procedure.
    */
   @Override
   @SuppressWarnings("unchecked")
   BaseDataType invoke(AsyncRequestImpl asyncReq,
                       character        name,
                       handle           h,
                       boolean          function,
                       boolean          dynamicFunction,
                       boolean          superCall,
                       boolean          transactionDistinct,
                       String           modes,
                       Object...        args)
   {
      if (asyncReq != null && asyncReq._isCancelled())
      {
         // if the procedure is cancelled, nothing else to do
         asyncReq.setResult(new RemoteInvocationResult());
         return new unknown();
      }

      if (superCall)
      {
         throw new IllegalArgumentException("superCall argument must always be false!");
      }

      if (h == null || !(h.get() instanceof PortTypeWrapper))
      {
         throw new IllegalArgumentException("The passed handle is not a port type procedure!");
      }

      if (modes == null || modes.length() != args.length)
      {
         throw new IllegalArgumentException("The parameter modes must always be specified!");
      }

      String ns                = wsdlDef.getTargetNamespace();
      String srvName           = service.getQName().getLocalPart();
      String operationName     = name.toStringMessage();
      PortTypeWrapper portProc = (PortTypeWrapper) h.get();
      String portTypeName      = portProc.name().toStringMessage();
      PortType portType        = wsdlDef.getPortType(new QName(ns, portTypeName));
      Operation targetOp       = null;
      
      // validate the operation name
      Iterator<Operation> itrOper = portType.getOperations().iterator();
      while (itrOper.hasNext())
      {
         Operation op = itrOper.next();
         // search is done case-insensitive
         if (op.getName().equalsIgnoreCase(operationName))
         {
            targetOp = op;
            break;
         }
      }
      if (targetOp == null)
      {
         final String msg =
          "Failure initializing SOAP Call: Cannot find a SOAP Operation named %s (11775) (11762)";
         ErrorManager.recordOrThrowError(11762, String.format(msg, operationName), false, true);
         
         return null;
      }
      
      Object[] inputParams = collectParameters(modes, args, true);
      Object[] outputParams = collectParameters(modes, args, false);

      Message inMsg = targetOp.getInput().getMessage();
      QName[][] inMsgParams = getMessageParameters(inMsg);
      
      Message outMsg = targetOp.getOutput().getMessage();
      QName[][] outMsgParams = getMessageParameters(outMsg);

      if (!validateParameters(targetOp, inputParams, outputParams,
                              inMsgParams.length, outMsgParams.length))
      {
         return null;
      }
      
      ConditionException condition = null;
      ServiceClient sc = null;

      // invoke the remote operation
      try
      {
         // create the message
         OMElement reqMsg = buildInputMessage(targetOp, inMsgParams, inputParams);
         if (reqMsg == null)
         {
            // request message could not be created, abort
            return null;
         }

         // invoke the request-header callback
         character requestHeader = invokeRequestHeaderCallback(portProc, targetOp);
         if (requestHeader == null)
         {
            // problems invoking the request-header callback, abort
            return null;
         }
         
         try
         {
            sc = new ServiceClient(null, wsdlDef, new QName(ns, srvName), portTypeName);
            if (!requestHeader.isUnknown())
            {
               // add the request header to the message
               try
               {
                  sc.addHeader(AXIOMUtil.stringToOM(requestHeader.toStringMessage()));
               }
               catch (XMLStreamException e)
               {
                  if (LOG.isLoggable(Level.SEVERE))
                  {
                     final String msg = "Could not add the SOAP Request header: %s";
                     LOG.log(Level.SEVERE, String.format(msg, requestHeader.toStringMessage()), e);
                  }
               }
            }
            
            if (asyncReq != null)
            {
               // set the request ID
               asyncReq.setRequestId(++nextAsyncReqId);
            }

            // TODO: how to terminate a running operation?
            //       maybe invoke it async? (note that at this time, if asyncReq is not null, we
            //       are already in a different thread, thus we still need to act as we are in 
            //       blocking mode.
            if (soapEndpoint != null)
            {
               // TODO: check if the endpoint is reachable now

               sc.getOptions().setTo(new EndpointReference(soapEndpoint));
            }
            
            // use the resolved operation name, not the received one, as AXIS is case-sensitive
            OMElement res = sc.sendReceive(new QName(ns, targetOp.getName()), reqMsg);
            
            // invoke the response-header callback
            if (portProc.getResponseHeaderProc() != null)
            {
               handle headerRef = new handle();
               SOAPHeaderImpl header = new SOAPHeaderImpl();
               headerRef.set(header);

               // collect the header entries
               MessageContext resp = sc.getLastOperationContext()
                                       .getMessageContext(WSDLConstants.MESSAGE_LABEL_IN_VALUE);
               SOAPHeader soapHeader = resp.getEnvelope().getHeader();
               if (soapHeader != null)
               {
                  Iterator hiter = soapHeader.getHeadersToProcess(null);
                  while (hiter.hasNext())
                  {
                     SOAPHeaderBlock respHeaderEntry= (SOAPHeaderBlock) hiter.next();
                     
                     SOAPHeaderEntryImpl entry = new SOAPHeaderEntryImpl();
                     // first register it as a header entry
                     header.addHeaderEntry(entry);
                     
                     entry.setSerialized(new longchar(respHeaderEntry.toString()));
                  }
               }
   
               // even if there is no header, an empty SOAP Header resource is sent.
               if (!invokeResponseHeaderCallback(portProc, headerRef, targetOp))
               {
                  // problems invoking the response-header callback, abort
                  return null;
               }
            }
   
            // process the response and set the output parameters
            if (!processOutputMessage(outMsgParams, outputParams, res))
            {
               return null;
            }
         }
         catch (AxisFault fault)
         {
            // the response-header callback is not invoked in case of  SOAP faults

            String faultActor  = (fault.getNodeURI() == null) ? "" : fault.getNodeURI();
            String faultString = (fault.getReason() == null) ? "" : fault.getReason();
            String faultCode   = 
               (fault.getFaultCode() == null) ? "" : fault.getFaultCode().getLocalPart();
            
            final String msg =
               "Web service operation %s generated a SOAP Fault. SOAP faultstring is: %s (11506)";
            int nlIdx = faultString.indexOf('\n');
            String faultMsg = (nlIdx >= 0) ? faultString.substring(0, nlIdx) : faultString;
            ErrorManager.recordOrThrowError(11506, String.format(msg, operationName, faultMsg),
                                            false, true);

            // if this is reached, we are in silent-error mode, thus generate the SOAP-fault object
            SOAPFaultDetailImpl soapFaultDetail = null;
            if (fault.getFaultDetailElement() != null)
            {
               String faultDetailXML = fault.getFaultDetailElement().toString();
               // use this to construct a SoapFaultDetail instance
               soapFaultDetail = SOAPFactory.createSOAPFaultDetail(faultDetailXML);
            }
            
            SOAPFaultImpl soapFault = 
               SOAPFactory.createSOAPFault(faultActor, faultCode, faultString, soapFaultDetail);

            ErrorManager.setErrorObjectDetail(soapFault);
         }
         
         // TODO: test how interruptions are treated when we are blocked invoking the operation
      }
      catch (ConditionException ce)
      {
         if (asyncReq == null)
         {
            // if not an async request, let it unwind
            throw ce;
         }

         condition = ce;
      }

      character res = new character();
      if (outputParams.length == 1)
      {
         res.assign((BaseDataType) outputParams[0]);
      }
      if (asyncReq != null)
      {
         RemoteInvocationResult asyncResult = new RemoteInvocationResult();
         asyncResult.setArguments(outputParams);
         asyncResult.setModes(StringHelper.repeatChar('I', outputParams.length));
         asyncResult.setResult(res);
         asyncResult.setError(condition);
         
         asyncReq.setResult(asyncResult);
      }
      
      return res;
   }
  
   /**
    * Validate the operation's parameters against the input and output parameters sent by the
    * caller.
    * <p>
    * The rules are:
    * <ol>
    *    <li>All output parameters must be a {@link BaseDataType}.</li>
    *    <li>If only one input (or output) parameter is specified, is assumed to be wrapped.</li>
    *    <li>If no input (or output) parameters are specified, then the operation must have no
    *        input (or output) parameters.</li>
    *    <li>
    * </ol>
    * @param    op
    *           The service operation to be invoked.
    * @param    input
    *           The passed input parameters.
    * @param    output
    *           The passed output parameters.
    * @param    inParams
    *           The number of input parameters defined at the operation.
    * @param    outParams
    *           The number of output parameters defined at the operation.
    *
    * @return   <code>true</code> if the parameters are valid.
    */
   private boolean validateParameters(Operation op,
                                      Object[]  input,
                                      Object[]  output,
                                      int       inParams,
                                      int       outParams)
   {
      // all outputParams must be BDT
      int i = 0;
      for (Object param : output)
      {
         if (!(param instanceof BaseDataType))
         {
            String msg = "Error invoking operation %s on service %s: Parameter %d is of " +
                         "type %s, which is not BaseDataType compatible!";
            msg = String.format(msg, op.getName(), service.getQName().getLocalPart(), i,
                                (param == null ? "null" : param.getClass().getName()));

            throw new IllegalStateException(msg);
         }
         i++;
      }
      
      // validate input parameters
      boolean paramError = false;
      if (input.length > 0)
      {
         if (input.length == 1)
         {
            // in this case, we assume that the parameter is passed as XML message OR there is
            // only one input parameter compatible with the one at the operation
            
            // type validation will be done later, when the input message is computed
         }
         else if (input.length != inParams)
         {
            paramError = true;
         }
      }
      else if (op.getInput().getMessage() != null && 
               !op.getInput().getMessage().getParts().isEmpty())
      {
         // no input parameters specified, fail as the operation has input parameters
         paramError = true;
      }
      
      // validate output parameters
      if (output.length > 0)
      {
         if (output.length == 1)
         {
            // in this case, we assume that the parameter is passed as XML message OR there is
            // only one output parameter compatible with the one at the operation
            
            // type validation will be done later, when the output message is computed
         }
         else if (output.length != outParams)
         {
            paramError = true;
         }
      }
      else if (op.getOutput().getMessage() != null && 
               !op.getOutput().getMessage().getParts().isEmpty())
      {
         // no output parameters specified, fail as the operation has output parameters
         paramError = true;
      }

      if (paramError)
      {
         String msg = "Failure initializing SOAP Call: Cannot match signature for operation '%s'";
         ErrorManager.recordOrThrowError(11762, String.format(msg, op.getName()));
      }
      
      return !paramError;
   }

   /**
    * Invoke the response-header callback, if it was set for the specified procedure.
    * <p>
    * This will also validate the response-header callback and the associated context procedure.
    * 
    * @param    portProc
    *           The {@link PortTypeWrapper port-type procedure}.
    * @param    soapHeader
    *           A handle pointing to the {@link SOAPHeader} resource.
    * @param    op
    *           The operation to be invoked.
    *           
    * @return   <code>true</code> if the callback was invoked and no error condition was raised.
    */
   private boolean invokeResponseHeaderCallback(PortTypeWrapper portProc,
                                                handle          soapHeader,
                                                Operation       op)
   {
      String procedure = portProc.getResponseHeaderProc();
      handle context = portProc.getResponseHeaderContext();
      
      if (procedure == null)
      {
         // nothing to invoke
         return false;
      }
      
      if (!context._isValid() || context.getResource() instanceof PortTypeWrapper)
      {
         final String msg1 = 
            "Could not process callback: invalid procedure handle for the callback. (11574)";
         final String msg2 = "Error receiving Web Service Response: An error occurred " +
            "processing SOAP Response Headers (11773)";
         
         final int[] errCodes = { 11574, 11773 };
         
         ErrorManager.recordOrThrowError(errCodes, new String[] {msg1, msg2}, false, true);
         
         return false;
      }
      
      String modes = "III";
      Element opElem = op.getDocumentationElement();
      character operNS = new character(opElem == null ? "" : opElem.getNamespaceURI());
      character operName = new character(op.getName());
      
      try
      {
         // validate the procedure name and parameters now. the modes are not validated.
         ArgValidationErrors argErr = 
            ControlFlowOps.reachableInternalEntry(context, procedure, false, modes, false,
                                                  soapHeader, operNS, operName);
         if (argErr == null || argErr != ArgValidationErrors.NO_ERROR)
         {
            showCallbackError(false, argErr, procedure);
            
            return false;
         }

         ControlFlowOps.invokeInWithMode(procedure, context, null, soapHeader, operNS, operName);
         return true;
      }
      catch (ErrorConditionException e)
      {
         // TODO: log it?

         final String msg =
            "Error receiving Web Service Response: An error occurred processing SOAP Response " +
            "Headers (11773)";
         ErrorManager.recordOrThrowError(11773, msg, false, true);
         
         return false;
      }
   }
   
   /**
    * Invoke the request-header callback, if it was set for the specified procedure.
    * <p>
    * This will also validate the request-header callback and the associated context procedure.
    * 
    * @param    portProc
    *           The {@link PortTypeWrapper port-type procedure}.
    * @param    op
    *           The operation to be invoked.
    *           
    * @return   an unknown {@link character} instance if nothing is needed to be invoked, 
    *           <code>null</code> in case of any raised error conditions or the request-header
    *           XML payload serialized as text, in case of success.
    */
   private character invokeRequestHeaderCallback(PortTypeWrapper portProc, Operation op)
   {
      String procedure = portProc.getRequestHeaderProc();
      handle context = portProc.getRequestHeaderContext();
      
      if (procedure == null)
      {
         // nothing to invoke
         return new character();
      }
      
      if (!context._isValid() || context.getResource() instanceof PortTypeWrapper)
      {
         final int[] errCodes = { 11574, 11767 };
         final String msg1 = 
            "Could not process callback: invalid procedure handle for the callback. (11574)";
         final String msg2 = "Error sending Web Service Request: An error occurred processing " +
            "SOAP Request Headers (11767)";
         
         ErrorManager.recordOrThrowError(errCodes, new String[] { msg1, msg2 }, false, true);
         
         return null;
      }

      String modes = "OIIO";
      handle soapHeader = new handle();
      Element opElem = op.getDocumentationElement();
      character operNS = new character(opElem == null ? "" : opElem.getNamespaceURI());
      character operName = new character(op.getName());
      logical deleteOnDone = new logical();
      String header = null;
      
      try
      {
         // validate the procedure name and parameters now. the modes are not validated.
         ArgValidationErrors argErr = 
            ControlFlowOps.reachableInternalEntry(context, procedure, false, modes, false,
                                                  soapHeader, operNS, operName, deleteOnDone);
         if (argErr == null || argErr != ArgValidationErrors.NO_ERROR)
         {
            showCallbackError(true, argErr, procedure);
            
            return null;
         }
         
         ControlFlowOps.invokeInWithMode(procedure, context, null,
                                         soapHeader, operNS, operName, deleteOnDone);
         
         if (soapHeader._isValid())
         {
            
            // no error is shown if the handle doesn't reference a SOAP-header...
            Object res = soapHeader.get();
            if (res instanceof SOAPHeaderImpl)
            {
               SOAPHeaderImpl hdr = (SOAPHeaderImpl) res;
               
               // get the header serialized to a string and return it.
               if (hdr.valid())
               {
                  header = hdr.asString();
               }
            }
            
            if (deleteOnDone.booleanValue())
            {
               // the soap header and the underlying DOM XML Nodes (refering the header entries)
               // get deleted
               HandleOps.delete(soapHeader);
            }
         }

         return new character(header);
      }
      catch (ErrorConditionException e)
      {
         final String msg =
            "Error sending Web Service Request: An error occurred processing SOAP Request " +
            "Headers (11767)";
         ErrorManager.recordOrThrowError(11767, msg, false, true);

         return null;
      }
   }
   
   /**
    * Show a callback error for a failed callback validation.
    * 
    * @param    request
    *           When <code>true</code>, the request-header callback is invoked.
    * @param    err
    *           The validation errors for this callback. <code>null</code> specifies the callback
    *           procedure was not found in the specified context.
    * @param    procedure
    *           The procedure name for this callback.
    */
   private void showCallbackError(boolean request, ArgValidationErrors err, String procedure)
   {
      final String reqMsg2 = "Error sending Web Service Request: An error occurred processing " +
                             "SOAP Request Headers (11767)";
      final String respMsg2 = "Error receiving Web Service Response: An error occurred " +
                              "processing SOAP Response Headers (11773)";

      final String msg2 = (request ? reqMsg2 : respMsg2);
      final int errCode2 = (request ? 11767 : 11773);
      
      final int errCode1;
      final String msg1;
      
      if (err != null)
      {
         if (err == ArgValidationErrors.INCORRECT_COUNT ||
             err == ArgValidationErrors.UNEXPECTED_PARAM)
         {
            msg1 = String.format("Callback error: mismatched number of parameters passed to " +
                                 "procedure %s. (10511)",
                                 procedure);
            errCode1 = 10511;
         }
         else
         {
            msg1 = String.format("%s-HEADER sent procedure %s mismatched parameters. (10512)",
                                 (request ? "REQUEST" : "RESPONSE"),
                                 procedure);
            errCode1 = 10512;
         }
      }
      else
      {
         msg1 = null;
         errCode1 = -1;
      }

      if (msg1 != null)
      {
         ErrorManager.recordOrThrowError(new int[] { errCode1, errCode2 },
                                         new String[] { msg1, msg2 },
                                         false, true);
      }
      else
      {
         ErrorManager.recordOrThrowError(errCode2, msg2, false, true);
      }
   }
   
   /**
    * Process the message received after an operation invocation and populate the output 
    * parameters sent by the caller accordingly.
    * <p>
    * If the caller passes only one input parameter of type {@link Text}, then it is assumed we 
    * are in wrapped mode and that parameter will hold the entire XML payload for the received
    * message, serialized as string.
    * 
    * @param    outMsgParams
    *           The (name, type) mappings of each parameter, defined at the operation's output.
    * @param    outputParams
    *           The output parameters sent by the caller; each element is a {@link BaseDataType}.
    * @param    message
    *           The message received after the operation invocation.
    *           
    * @return   <code>true</code> if message processed or <code>false</code> if there were 
    *           problems during conversion of the received to their {@link BaseDataType} expected
    *           type.
    */
   @SuppressWarnings("unchecked")
   private boolean processOutputMessage(QName[][] outMsgParams,
                                        Object[]  outputParams,
                                        OMElement message)
   {
      if (outputParams.length == 1)
      {
         // check if we are emitting the entire XML
         BaseDataType param = (BaseDataType) outputParams[0];
         if (param instanceof Text)
         {
            // TODO: what if there is only one message output parameter, of type char? should we
            // unwrap it?
            
            Text txt = (Text) param;
            txt.assign(message.toString());
            
            return true;
         }
      }
      
      // in all other cases, match the output parameters with the received message

      int i = 0;
      Iterator<?> itrElem = message.getChildren();
      while (itrElem.hasNext())
      {
         Object child = itrElem.next();
         if (!(child instanceof OMElement))
         {
            continue;
         }
         
         OMElement omChild = (OMElement) child;
         BaseDataType param = (BaseDataType) outputParams[i];
         
         BaseDataType value = unmarshallValue(outMsgParams[i][0], outMsgParams[i][1], omChild);
         if (value == null)
         {
            String xmlType = outMsgParams[i][1].getLocalPart();
            String legacyType = param.getTypeName();
            
            final String msg = "Unable to convert XML data type %s to 4GL type %s  (11793)";
            ErrorManager.recordOrThrowError(11793, String.format(msg, xmlType, legacyType));
            
            return false;
         }
         
         param.assign(value);
         
         i++;
      }
      
      return true;
   }
   
   /**
    * Using the input parameters configured at the operation's input message and the input 
    * parameters sent by the caller, create a new XML payload representing the input message
    * passed to the operation call.
    * <p>
    * The caller may not pass null or unknown values.  If the caller passes only one input 
    * parameter of type {@link Text} or {@link String}, then it is assumed we are in wrapped 
    * mode and the parameter holds the entire XML payload.
    * 
    * @param    op
    *           The operation to be invoked.
    * @param    inMsgParams
    *           The (name, type) mappings of each parameter, defined at the operation's input.
    * @param    inputParams
    *           The input parameters sent by the caller.
    *           
    * @return   A new {@link OMElement} structure with the XML payload for the input message.
    */
   @SuppressWarnings("unchecked")
   private OMElement buildInputMessage(Operation op, QName[][] inMsgParams, Object[] inputParams)
   {
      BaseDataType[] params = new BaseDataType[inputParams.length];
      for (int i = 0; i < inputParams.length; i++)
      {
         Object param = inputParams[i];
         
         if (param == null ||
             (param instanceof BaseDataType && ((BaseDataType) param).isUnknown()))
         {
            ErrorManager.recordOrThrowError(11790, "The 4GL value may not be UNKNOWN (11790)",
                                            false, true);
            return null;
         }
         
         if (param instanceof BaseDataType)
         {
            params[i] = (BaseDataType) param;
         }
         else
         {
            // TODO: build a BDT based on the param's java type
            params[i] = new character(inputParams[i].toString());
         }
      }
      
      if (params.length == 1)
      {
         // if there is only one input parameter, then that parameter must be a character/longchar
         // with the input message encoded as XML.
         
         // get the string representation of the parameter
         Object oxml = params[0];
         String xml = null;
         if (oxml instanceof Text)
         {
            xml = ((Text) oxml).toStringMessage();
         }
         else if (oxml instanceof String)
         {
            xml = (String) oxml;
         }
         else
         {
            xml = (oxml == null ? "?" : oxml.toString());
         }
         
         // attempt to parse the input parameter to an XML
         try
         {
            // TODO: the exception to this is when the only argument is of type char
            // in this case, it is possible to pass it in unwrapped mode

            // 4GL does not care of what the XML contains - it must be a valid XML
            return AXIOMUtil.stringToOM(xml);
         }
         catch (XMLStreamException e)
         {
            final String msg = "Malformed XML fragment:  %s (11781)";
            ErrorManager.recordOrThrowError(11781, String.format(msg, e.getMessage()),
                                            false, true);
            
            return null;
         }
      }
      
      // we need to build a new request message
      String ns     = wsdlDef.getTargetNamespace();
      String prefix = wsdlDef.getPrefix(ns);
      
      OMElement root = soapFactory.createOMElement(new QName(ns, op.getName(), prefix));
      for (int i = 0; i < params.length; i++)
      {
         // the order of the passed parameters must match the order of the parts at the message
         // definition. the check is done by converting the value to the expected type
         
         // attempt to convert to the XSD type, and from there to string
         String sval = marshallValue(inMsgParams[i][0], inMsgParams[i][1], params[i]);
         if (sval == null)
         {
            // TODO: show error
            return null;
         }
         
         QName elName = new QName(ns, inMsgParams[i][0].getLocalPart(), prefix);
         OMElement child = soapFactory.createOMElement(elName);
         child.addChild(soapFactory.createOMText(child, sval));
         root.addChild(child);
      }

      return root;
   }
   
   /**
    * Resolve the parameters for the given message.
    * 
    * @param    msg
    *           The message.
    *           
    * @return   An array where, for each element, on index 0 is the parameter's name and on index
    *           1 is the parameter's type.
    */
   @SuppressWarnings("unchecked")
   private QName[][] getMessageParameters(Message msg)
   {
      buildSchema();

      // TODO: complex types can define parameters as i.e. "sequence", "all" or "choice". this 
      // means is not enough to determine the entire list of possible parameters, as for "all" and
      // "choice" matching might be done via the parameter's type (for input) or the passed name
      // and type (for output).  thus, this method might need to return some complex structure
      // instead of (name, type) mappings.
      
      QName msgName = msg.getQName();
      if (msgParams.containsKey(msgName))
      {
         return msgParams.get(msgName);
      }
      
      // resolve the parameters
      List<QName[]> params = new ArrayList<>();
      
      Iterator<Part> iter = msg.getParts().values().iterator();
      while (iter.hasNext())
      {
         Part part = iter.next();
         
         QName name = part.getElementName();
         
         /// walk the schemas and determine the element
         for (XmlSchema schema : xmlSchemas)
         {
            XmlSchemaElement element = schema.getElementByName(name);
            
            resolveParameters(element, name, params);
         }
      }
      
      return params.toArray(new QName[][] {});
   }
   
   /**
    * If the {@link #xmlSchemas} list is not yet build, extract all the schema extensibility 
    * elements from the {@link #wsdlDef WSDL definition} and save them in the {@link #xmlSchemas}
    * list.
    */
   @SuppressWarnings("unchecked")
   private void buildSchema()
   {
      if (xmlSchemas != null)
      {
         return;
      }

      xmlSchemas = new ArrayList<>();

      Types types = wsdlDef.getTypes();
      Iterator<ExtensibilityElement> iter = types.getExtensibilityElements().iterator();
      while (iter.hasNext())
      {
         ExtensibilityElement el = iter.next();
         if (el instanceof Schema)
         {
            Schema schema = (Schema) el;
            String baseUri = schema.getDocumentBaseURI();

            XmlSchemaCollection xsc = new XmlSchemaCollection();
            if (baseUri != null)
            {
               xsc.setBaseUri(baseUri);
            }
            XmlSchema xmlSchema = xsc.read(schema.getElement());
            
            xmlSchemas.add(xmlSchema);
         }
      }
   }

   /**
    * Resolve the parameters from the specified schema element and save them in the 
    * <code>params</code> list: on index 0 the name and on index 1 the schema type.
    * 
    * 
    * @param    element
    *           The schema element which needs to be resolved.
    * @param    name
    *           The name of this element.
    * @param    params
    *           The list where to save the parameters. 
    */
   private void resolveParameters(XmlSchemaElement element, QName name, List<QName[]> params)
   {
      if (element == null)
      {
         return;
      }

      // TODO: what other features need to be treated?

      XmlSchemaType elType = element.getSchemaType();
      if (elType instanceof XmlSchemaSimpleType)
      {
         // found a simple type, use it
         params.add(new QName[] { name, element.getSchemaTypeName() });
      }
      else if (elType instanceof XmlSchemaComplexType)
      {
         XmlSchemaComplexType type = (XmlSchemaComplexType) elType;
         XmlSchemaParticle particle = type.getParticle();
         if (particle instanceof XmlSchemaAny)
         {
            // TODO:
         }
         else if (particle instanceof XmlSchemaElement)
         {
            // TODO:
         }
         else if (particle instanceof XmlSchemaAll)
         {
            // TODO:
         }
         else if (particle instanceof XmlSchemaChoice)
         {
            // TODO:
         }
         else if (particle instanceof XmlSchemaSequence)
         {
            // TODO:
            XmlSchemaSequence xmlSeq = (XmlSchemaSequence) particle;
            int no = xmlSeq.getItems().getCount();
            for (int i = 0; i < no; i++)
            {
               XmlSchemaObject obj = xmlSeq.getItems().getItem(i);
               if (obj instanceof XmlSchemaElement)
               {
                  XmlSchemaElement elObj = (XmlSchemaElement) obj;
                  resolveParameters(elObj, elObj.getQName(), params);
               }
            }
         }
         else if (particle instanceof XmlSchemaGroupRef)
         {
            // TODO:
         }
         
      }
   }
   
   /**
    * Unmarshall a value received as a response for an service operation request to a
    * {@link BaseDataType}, depending on the parameter's type defined at the operation's output
    * message.  This will also validate the value with the parameter's type defined at the schema.
    * 
    * @param    paramName
    *           The parameter's name, as defined at the SOAP message.
    * @param    paramType
    *           The parameter's type, as defined at the schema.
    * @param    param
    *           The response received after an operation invocation.
    *           
    * @return   A {@link BaseDataType} instance for the received parameter or <code>null</code> if
    *           conversion is not possible.
    */
   private BaseDataType unmarshallValue(QName paramName, QName paramType, OMElement param)
   {
      // TODO:
      return null;
   }
   
   /**
    * Marshall a value from Java/Progress style to SOAP style.  This will also validate the input
    * value with the parameter's type defined at the schema.
    * 
    * @param    paramName
    *           The parameter's name, as defined at the SOAP message.
    * @param    paramType
    *           The parameter's type, as defined at the schema.
    * @param    value
    *           The JAva or Progress-style value to be converted.
    *           
    * @return   The string representation of this value or <code>null</code> if marshalling is not
    *           possible (due to type conversion incompatibility).
    */
   private String marshallValue(QName paramName, QName paramType, Object value)
   {
      // TODO:
      return null;
   }
   
   /**
    * Search the given parameter modes and return the input or output parameters, based on the
    * <code>input</code> flag.
    * <p>
    * The {@link SourceNameMapper#INPUT_OUTPUT_MODE INPUT-OUTPUT} will be returned in both cases.
    *  
    * @param    modes
    *           The parameter modes.
    * @param    args
    *           The passed arguments.
    * @param    input
    *           On <code>true</code>, collect input parameters; else, collect output parameters.
    *           
    * @return   The collected parameters from the <code>args</code> array.
    */
   private Object[] collectParameters(String modes, Object[] args, boolean input)
   {
      List<Object> params = new ArrayList<>();
      for (int i = 0; i < modes.length(); i++)
      {
         char mode = modes.charAt(i);
         
         if (mode == SourceNameMapper.INPUT_OUTPUT_MODE ||
             (input && mode == SourceNameMapper.INPUT_MODE) ||
             (!input && mode == SourceNameMapper.OUTPUT_MODE))
         {
            params.add(args[i]);
         }
      }
      
      return params.toArray();
   }
   
   /**
    * Container for various web-service connect options. 
    */
   public static class WebServiceConnectOptions
   {
      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_WSDL} option. */
      private String wsdlDocument;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_WSDL_USERID} option. */
      private String wsdlUserid;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_WSDL_PASSWORD} option. */
      private String wsdlPassword;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_SERVICE} option. */
      private String service;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_SERVICE_NAMESPACE} option. */
      private String serviceNamespace;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_PORT} option. */
      private String port;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_BINDING} option. */
      private String binding;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_BINDING_NAMESPACE} option. */
      private String bindingNamespace;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_SOAP_ENDPOINT} option. */
      private String soapEndpoint;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_SOAP_ENDPOINT_USER} option. */
      private String soapEndpointUserid;

      /**
       * The value of the {@link WebServiceHelper#CONNECT_OPTION_SOAP_ENDPOINT_PASSWORD} option.
       */
      private String soapEndpointPassword;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_TARGET_NAMESPACE} option. */
      private String targetNamespace;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_MAX_CONNECTIONS} option. */
      private int maxConnections;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_CONNECTION_LIFETIME} option. */
      private int connectionLifetime;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_NO_SESSION_REUSE} option. */
      private boolean noSessionReuse;

      /** The value of the {@link WebServiceHelper#CONNECT_OPTION_NO_HOST_VERIFY} option. */
      private boolean noHostVerify;
      
      public WebServiceConnectOptions(String wsdlDocument)
      {
         this.wsdlDocument = wsdlDocument;
      }

      public String getWsdlDocument()
      {
         return wsdlDocument;
      }

      public String getWsdlUserid()
      {
         return wsdlUserid;
      }

      public void setWsdlUserid(String wsdlUserid)
      {
         this.wsdlUserid = wsdlUserid;
      }

      public String getWsdlPassword()
      {
         return wsdlPassword;
      }

      public void setWsdlPassword(String wsdlPassword)
      {
         this.wsdlPassword = wsdlPassword;
      }

      public String getService()
      {
         return service;
      }

      public void setService(String service)
      {
         this.service = service;
      }

      public String getServiceNamespace()
      {
         return serviceNamespace;
      }

      public void setServiceNamespace(String serviceNamespace)
      {
         this.serviceNamespace = serviceNamespace;
      }

      public String getPort()
      {
         return port;
      }

      public void setPort(String port)
      {
         this.port = port;
      }

      public String getBinding()
      {
         return binding;
      }

      public void setBinding(String binding)
      {
         this.binding = binding;
      }

      public String getBindingNamespace()
      {
         return bindingNamespace;
      }

      public void setBindingNamespace(String bindingNamespace)
      {
         this.bindingNamespace = bindingNamespace;
      }

      public String getSoapEndpoint()
      {
         return soapEndpoint;
      }

      public void setSoapEndpoint(String soapEndpoint)
      {
         this.soapEndpoint = soapEndpoint;
      }

      public String getSoapEndpointUserid()
      {
         return soapEndpointUserid;
      }

      public void setSoapEndpointUserid(String soapEndpointUserid)
      {
         this.soapEndpointUserid = soapEndpointUserid;
      }

      public String getSoapEndpointPassword()
      {
         return soapEndpointPassword;
      }

      public void setSoapEndpointPassword(String soapEndpointPassword)
      {
         this.soapEndpointPassword = soapEndpointPassword;
      }

      public String getTargetNamespace()
      {
         return targetNamespace;
      }

      public void setTargetNamespace(String targetNamespace)
      {
         this.targetNamespace = targetNamespace;
      }

      public int getMaxConnections()
      {
         return maxConnections;
      }

      public void setMaxConnections(int maxConnections)
      {
         this.maxConnections = maxConnections;
      }

      public int getConnectionLifetime()
      {
         return connectionLifetime;
      }

      public void setConnectionLifetime(int connectionLifetime)
      {
         this.connectionLifetime = connectionLifetime;
      }

      public boolean isNoSessionReuse()
      {
         return noSessionReuse;
      }

      public void setNoSessionReuse(boolean noSessionReuse)
      {
         this.noSessionReuse = noSessionReuse;
      }

      public boolean isNoHostVerify()
      {
         return noHostVerify;
      }

      public void setNoHostVerify(boolean noHostVerify)
      {
         this.noHostVerify = noHostVerify;
      }
   }
}
